import asyncio
import logging

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, final

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from aiosalesforce.client import Salesforce


class Auth(ABC):
    """
    Base class for Salesforce authentication.

    """

    def __init__(self) -> None:
        self.__access_token: str | None = None
        self.__lock = asyncio.Lock()

    @final
    async def get_access_token(self, client: "Salesforce") -> str:
        """
        Get access token.

        If this is the first time this method is called, it will acquire a new
        access token from Salesforce.

        Parameters
        ----------
        client : Salesforce
            Salesforce client.

        Returns
        -------
        str
            Access token

        """
        async with self.__lock:
            if self.__access_token is None:
                logger.debug(
                    "Acquiring new access token using %s for %s",
                    self.__class__.__name__,
                    client.base_url,
                )
                self.__access_token = await self._acquire_new_access_token(client)
            elif self.expired:
                logger.debug(
                    "Token expired, refreshing access token using %s for %s",
                    self.__class__.__name__,
                    client.base_url,
                )
                self.__access_token = await self._refresh_access_token(client)
            return self.__access_token

    @final
    async def refresh_access_token(self, client: "Salesforce") -> str:
        """
        Refresh the access token.

        Parameters
        ----------
        client : Salesforce
            Salesforce client.

        Returns
        -------
        str
            Access token

        """
        if self.__access_token is None:
            raise RuntimeError("No access token to refresh")
        token_before_refresh = self.__access_token
        async with self.__lock:
            if self.__access_token == token_before_refresh:
                logger.debug(
                    "Refreshing access token using %s for %s",
                    self.__class__.__name__,
                    client.base_url,
                )
                self.__access_token = await self._refresh_access_token(client)
            return self.__access_token

    @abstractmethod
    async def _acquire_new_access_token(self, client: "Salesforce") -> str:
        """
        Acquire a new access token from Salesforce.

        Implementation is responsible for emitting RequestEvent and ResponseEvent.

        Parameters
        ----------
        client : Salesforce
            Salesforce client.

        Returns
        -------
        str
            Access token

        """

    async def _refresh_access_token(self, client: "Salesforce") -> str:
        """
        Refresh the access token.

        Implementation is responsible for emitting RequestEvent and ResponseEvent.

        Parameters
        ----------
        client : Salesforce
            Salesforce client.

        Returns
        -------
        str
            Access token

        """
        return await self._acquire_new_access_token(client)

    @property
    def expired(self) -> bool:
        """True if the access token is expired."""
        if self.__access_token is None:  # pragma: no cover
            raise RuntimeError("Cannot check expiration of a non-existent access token")
        # By default, assumes the access token never expires
        # Salesforce client automatically refreshes the token after 401 response
        return False



import re

from dataclasses import dataclass
from functools import cached_property
from typing import Literal

from httpx import Request, Response


class ResponseMixin:
    """Mixin class providing properties for events which may have response."""

    response: Response | None

    @property
    def consumed(self) -> int | None:
        """Number of API calls consumed in the current 24-hour period."""
        return self.__api_usage[0]

    @property
    def remaining(self) -> int | None:
        """Number of API calls remaining in the current 24-hour period."""
        return self.__api_usage[1]

    @cached_property
    def __api_usage(self) -> tuple[int, int] | tuple[None, None]:
        if self.response is None:
            return (None, None)
        try:
            match_ = re.fullmatch(
                r"^api-usage=(\d+)/(\d+)$",
                str(self.response.headers["Sforce-Limit-Info"]).strip(),
            )
        except KeyError:
            return (None, None)
        if match_ is None:  # pragma: no cover
            return (None, None)
        consumed, remaining = match_.groups()
        return int(consumed), int(remaining)


@dataclass
class Event:
    """Base class for all events."""

    type: Literal[
        "request",
        "response",
        "retry",
        "rest_api_call_consumption",
        "bulk_api_batch_consumption",
    ]


@dataclass
class RequestEvent(Event):
    """Emitted before a request is sent for the first time."""

    type: Literal["request"]
    request: Request


@dataclass
class RetryEvent(Event, ResponseMixin):
    """Emitted immediately before a request is retried."""

    type: Literal["retry"]
    attempt: int
    request: Request
    response: Response | None = None
    exception: Exception | None = None


@dataclass
class ResponseEvent(Event, ResponseMixin):
    """Emitted after an OK (status code < 300) response is received."""

    type: Literal["response"]
    response: Response


@dataclass
class RestApiCallConsumptionEvent(Event, ResponseMixin):
    """Emitted after a REST API call is consumed."""

    type: Literal["rest_api_call_consumption"]
    response: Response
    count: int


@dataclass
class BulkApiBatchConsumptionEvent(Event, ResponseMixin):
    """Emitted after a Bulk API batch is consumed."""

    type: Literal["bulk_api_batch_consumption"]
    response: Response
    count: int



import asyncio
import logging
import re
import warnings

from functools import cached_property, wraps
from typing import Any, AsyncIterator, Awaitable, Callable, Iterable, NoReturn

import httpx

from aiosalesforce import __version__
from aiosalesforce.auth import Auth
from aiosalesforce.bulk import BulkClientV2
from aiosalesforce.events import (
    Event,
    EventBus,
    RequestEvent,
    ResponseEvent,
)
from aiosalesforce.exceptions import SalesforceWarning, raise_salesforce_error
from aiosalesforce.retries import POLICY_DEFAULT, RetryPolicy
from aiosalesforce.sobject import SobjectClient
from aiosalesforce.utils import json_loads

logger = logging.getLogger(__name__)


class Salesforce:
    """
    Salesforce API client.

    Parameters
    ----------
    httpx_client : httpx.AsyncClient
        HTTP client.
    base_url : str
        Base URL of the Salesforce instance.
        Must be in the format:\n
        * Production    : https://[MyDomainName].my.salesforce.com
        * Sandbox       : https://[MyDomainName]-[SandboxName].sandbox.my.salesforce.com
        * Developer org : https://[MyDomainName].develop.my.salesforce.com\n
    auth : Auth
        Authentication object.
    version : str, optional
        Salesforce API version.
        By default, uses the latest version.
    event_hooks : Iterable[Callable[[Event], Awaitable[None] | None]], optional
        Functions or coroutines executed when an event occurs.
        Hooks are executed concurrently and order of execution is not guaranteed.
        All hooks must be thread-safe.
    retry_policy : RetryPolicy, optional
        Retry policy for requests.
        The default policy retries requests up to 3 times with exponential backoff
        and retries the following:\n
        * httpx Transport errors (excluding timeouts)
        * Server errors (5xx)
        * Row lock errors
        * Rate limit errors\n
        Set to None to disable retries.
    concurrency_limit : int, optional
        Maximum number of simultaneous requests to Salesforce.
        The default is 100.

    """

    httpx_client: httpx.AsyncClient
    auth: Auth
    event_bus: EventBus
    retry_policy: RetryPolicy
    _semaphore: asyncio.Semaphore

    def __init__(
        self,
        httpx_client: httpx.AsyncClient,
        base_url: str,
        auth: Auth,
        version: str = "60.0",
        event_hooks: Iterable[Callable[[Event], Awaitable[None] | None]] | None = None,
        retry_policy: RetryPolicy | None = POLICY_DEFAULT,
        concurrency_limit: int = 100,
    ) -> None:
        self.httpx_client = httpx_client
        self.base_url = base_url
        self.auth = auth
        self.version = version

        self.event_bus = EventBus(event_hooks)
        self.retry_policy = retry_policy or RetryPolicy()
        self._semaphore = asyncio.Semaphore(concurrency_limit)

    @property
    def version(self) -> str:
        return self.__version

    @version.setter
    def version(self, value: str) -> None:
        """API version in the format '60.0'."""
        if not (match_ := re.fullmatch(r"^(v)?(\d+)(\.(0)?)?$", value)):
            raise ValueError(
                f"Invalid Salesforce API version: '{value}'. "
                f"A valid version should look like '60.0'."
            )
        self.__version = f"{match_.groups()[1]}.0"

    @property
    def base_url(self) -> str:
        """Base URL in the format https://[subdomain(s)].my.salesforce.com"""
        return self.__base_url

    @base_url.setter
    def base_url(self, value: str) -> None:
        match_ = re.fullmatch(
            r"(https://[a-zA-Z0-9-]+(\.(sandbox|develop))?\.my\.salesforce\.com).*",
            value.strip(" ").lower(),
        )
        if not match_:
            raise ValueError(
                "\n".join(
                    [
                        f"Invalid Salesforce URL: {value}",
                        "Supported formats:",
                        "  Production    : https://[MyDomainName].my.salesforce.com",
                        "  Sandbox       : https://[MyDomainName]-[SandboxName].sandbox.my.salesforce.com",
                        "  Developer org : https://[MyDomainName].develop.my.salesforce.com",
                    ]
                )
            )
        self.__base_url = str(match_.groups()[0])

    @wraps(httpx.AsyncClient.request)
    async def request(self, *args, **kwargs) -> httpx.Response:
        """
        Make an HTTP request to Salesforce.

        """
        request = self.httpx_client.build_request(*args, **kwargs)
        access_token = await self.auth.get_access_token(self)
        request.headers.update(
            {
                "Authorization": f"Bearer {access_token}",
                "User-Agent": f"aiosalesforce/{__version__}",
                "Sforce-Call-Options": f"client=aiosalesforce/{__version__}",
                "Sforce-Line-Ending": "LF",
            }
        )
        await self.event_bus.publish_event(
            RequestEvent(type="request", request=request)
        )
        retry_context = self.retry_policy.create_context()
        response = await retry_context.send_request_with_retries(
            httpx_client=self.httpx_client,
            event_bus=self.event_bus,
            semaphore=self._semaphore,
            request=request,
        )
        if response.status_code == 401:
            access_token = await self.auth.refresh_access_token(self)
            request.headers["Authorization"] = f"Bearer {access_token}"
            response = await retry_context.send_request_with_retries(
                httpx_client=self.httpx_client,
                event_bus=self.event_bus,
                semaphore=self._semaphore,
                request=request,
            )
        if not response.is_success:
            raise_salesforce_error(response)
        if "Warning" in response.headers:
            warnings.warn(response.headers["Warning"], SalesforceWarning)
        await self.event_bus.publish_event(
            ResponseEvent(type="response", response=response)
        )
        return response

    async def get_limits(self) -> dict[str, Any]:
        """
        Get Salesforce org limits.

        Returns
        -------
        dict
            Salesforce org limits.

        """
        response = await self.request(
            "GET",
            f"{self.base_url}/services/data/v{self.version}/limits",
            headers={"Accept": "application/json"},
        )
        return json_loads(response.content)

    async def query(
        self,
        query: str,
        include_all_records: bool = False,
    ) -> AsyncIterator[dict]:
        """
        Execute a SOQL query.

        Parameters
        ----------
        query : str
            SOQL query.
        include_all_records : bool, default False
            If True, includes all (active/deleted/archived) records.

        Yields
        -------
        dict
            Query result record.

        """
        operation = "query" if not include_all_records else "queryAll"

        next_url: str | None = None
        while True:
            if next_url is None:
                response = await self.request(
                    "GET",
                    f"{self.base_url}/services/data/v{self.version}/{operation}",
                    params={"q": query},
                    headers={"Accept": "application/json"},
                )
            else:
                response = await self.request(
                    "GET",
                    f"{self.base_url}{next_url}",
                    headers={"Accept": "application/json"},
                )
            response_json: dict = json_loads(response.content)
            for record in response_json["records"]:
                yield record
            next_url = response_json.get("nextRecordsUrl", None)
            if next_url is None:
                break

    @cached_property
    def sobject(self) -> SobjectClient:
        """
        Salesforce REST API sObject client.

        Use this client to perform CRUD operations on individual sObjects.

        """
        return SobjectClient(self)

    @cached_property
    def bulk_v1(self) -> NoReturn:
        """
        Get Salesforce Bulk API 1.0 client.

        Use this client to execute bulk ingest and query operations.

        """
        raise NotImplementedError("Bulk API v1 is currently not supported")

    @cached_property
    def bulk_v2(self) -> BulkClientV2:
        """
        Get Salesforce Bulk API 2.0 client.

        Use this client to execute bulk ingest and query operations.

        """
        return BulkClientV2(self)



__all__ = [
    "Auth",
    "ClientCredentialsFlow",
    "JwtBearerFlow",
    "SoapLogin",
]

from .base import Auth
from .client_credentials_flow import ClientCredentialsFlow
from .jwt_bearer_flow import JwtBearerFlow
from .soap import SoapLogin



import re
import textwrap
import time

from typing import TYPE_CHECKING

from aiosalesforce.events import (
    RequestEvent,
    ResponseEvent,
)
from aiosalesforce.exceptions import AuthenticationError

from .base import Auth

if TYPE_CHECKING:
    from aiosalesforce.client import Salesforce


class SoapLogin(Auth):
    """
    Authenticate using the SOAP login method.

    https://developer.salesforce.com/docs/atlas.en-us.api.meta/api/sforce_api_calls_login.htm

    Parameters
    ----------
    username : str
        Username.
    password : str
        Password.
    security_token : str
        Security token.

    """

    def __init__(
        self,
        username: str,
        password: str,
        security_token: str,
    ):
        super().__init__()
        self.username = username
        self.password = password
        self.security_token = security_token

        self._expiration_time: float | None = None

    async def _acquire_new_access_token(self, client: "Salesforce") -> str:
        soap_xml_payload = f"""
        <?xml version="1.0" encoding="utf-8" ?>
        <env:Envelope
            xmlns:xsd="http://www.w3.org/2001/XMLSchema"
            xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
            xmlns:env="http://schemas.xmlsoap.org/soap/envelope/">
            <env:Body>
                <n1:login xmlns:n1="urn:partner.soap.sforce.com">
                    <n1:username>{self.username}</n1:username>
                    <n1:password>{self.password}{self.security_token}</n1:password>
                </n1:login>
            </env:Body>
        </env:Envelope>
        """
        request = client.httpx_client.build_request(
            "POST",
            f"{client.base_url}/services/Soap/u/{client.version}",
            content=textwrap.dedent(soap_xml_payload).strip(),
            headers={
                "Content-Type": "text/xml; charset=UTF-8",
                "SOAPAction": "login",
                "Accept": "text/xml",
            },
        )
        await client.event_bus.publish_event(
            RequestEvent(
                type="request",
                request=request,
            )
        )
        retry_context = client.retry_policy.create_context()
        response = await retry_context.send_request_with_retries(
            httpx_client=client.httpx_client,
            event_bus=client.event_bus,
            semaphore=client._semaphore,
            request=request,
        )
        response_text = response.text
        if not response.is_success:
            try:
                exception_code = str(
                    re.search(
                        r"<sf:exceptionCode>(.+)<\/sf:exceptionCode>",
                        response_text,
                    ).groups()[0]  # type: ignore
                )
            except AttributeError:  # pragma: no cover
                exception_code = None
            try:
                exception_message = str(
                    re.search(
                        r"<sf:exceptionMessage>(.+)<\/sf:exceptionMessage>",
                        response_text,
                    ).groups()[0]  # type: ignore
                )
            except AttributeError:  # pragma: no cover
                exception_message = response_text
            raise AuthenticationError(
                message=(
                    f"[{exception_code}] {exception_message}"
                    if exception_code
                    else exception_message
                ),
                response=response,
                error_code=exception_code,
                error_message=exception_message,
            )
        match_ = re.search(r"<sessionId>(.+)<\/sessionId>", response_text)
        if match_ is None:  # pragma: no cover
            raise AuthenticationError(
                f"Failed to parse sessionId from the SOAP response: {response_text}",
                response,
            )
        session_id = match_.groups()[0]

        # Parse expiration time
        match_ = re.search(
            r"<sessionSecondsValid>(.+)<\/sessionSecondsValid>",
            response_text,
        )
        self._expiration_time = None
        if match_ is not None:
            try:
                self._expiration_time = time.time() + int(match_.groups()[0])
            except ValueError:  # pragma: no cover
                pass

        await client.event_bus.publish_event(
            ResponseEvent(
                type="response",
                response=response,
            )
        )
        return session_id

    @property
    def expired(self) -> bool:
        super().expired
        if self._expiration_time is None:  # pragma: no cover
            return False
        return self._expiration_time <= time.time()



import pathlib
import time

from typing import TYPE_CHECKING

try:
    import jwt

    from cryptography.hazmat.primitives import serialization
except ImportError:  # pragma: no cover
    jwt = None  # type: ignore
    serialization = None  # type: ignore

from aiosalesforce.auth.base import Auth
from aiosalesforce.events import RequestEvent, ResponseEvent
from aiosalesforce.exceptions import AuthenticationError
from aiosalesforce.utils import json_loads

if TYPE_CHECKING:
    from aiosalesforce.client import Salesforce


class JwtBearerFlow(Auth):
    """
    Authenticate using the OAuth 2.0 JWT Bearer Flow.

    https://help.salesforce.com/s/articleView?id=sf.remoteaccess_oauth_jwt_flow.htm&type=5

    Parameters
    ----------
    client_id : str
        Client ID.
    username: str
        Username.
    private_key_file: str | pathlib.Path
        Path to private key file.
    private_key_passphrase: str, optional
        Passphrase for private key file.
        By default assumed to be unencrypted.
    timeout : float, optional
        Timeout for the access token in seconds.
        By default assumed to never expire.

    """

    def __init__(
        self,
        client_id: str,
        username: str,
        private_key_file: str | pathlib.Path,
        private_key_passphrase: str | None = None,
        timeout: float | None = None,
    ) -> None:
        super().__init__()
        self.client_id = client_id
        self.username = username
        self.private_key_file = private_key_file
        self.private_key_passphrase = private_key_passphrase
        self.timeout = timeout

        self._expiration_time: float | None = None

        if jwt is None or serialization is None:  # pragma: no cover
            raise ImportError("Install aiosalesforce[jwt] to use the JwtBearerFlow.")

    async def _acquire_new_access_token(self, client: "Salesforce") -> str:
        payload = {
            "iss": self.client_id,
            "aud": "https://test.salesforce.com"
            if client.base_url.endswith(".sandbox.my.salesforce.com")
            else "https://login.salesforce.com",
            "sub": self.username,
            "exp": int(time.time()) + 300,
        }
        with open(self.private_key_file, "rb") as file:
            private_key = serialization.load_pem_private_key(
                data=file.read(),
                password=self.private_key_passphrase.encode("utf-8")
                if self.private_key_passphrase is not None
                else None,
            )
        assertion = jwt.encode(
            payload,
            private_key,  # type: ignore
            algorithm="RS256",
            headers={"alg": "RS256"},
        )
        request = client.httpx_client.build_request(
            "POST",
            f"{client.base_url}/services/oauth2/token",
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "Accept": "application/json",
            },
            data={
                "grant_type": "urn:ietf:params:oauth:grant-type:jwt-bearer",
                "assertion": assertion,
            },
        )
        await client.event_bus.publish_event(
            RequestEvent(
                type="request",
                request=request,
            )
        )
        retry_context = client.retry_policy.create_context()
        response = await retry_context.send_request_with_retries(
            httpx_client=client.httpx_client,
            event_bus=client.event_bus,
            semaphore=client._semaphore,
            request=request,
        )
        if not response.is_success:
            try:
                response_json = json_loads(response.content)
                error_code = response_json["error"]
                error_message = response_json["error_description"]
            except Exception:  # pragma: no cover
                error_code = None
                error_message = response.text
            raise AuthenticationError(
                f"[{error_code}] {error_message}" if error_code else error_message,
                response=response,
                error_code=error_code,
                error_message=error_message,
            )
        await client.event_bus.publish_event(
            ResponseEvent(
                type="response",
                response=response,
            )
        )
        if self.timeout is not None:
            self._expiration_time = time.time() + self.timeout
        return json_loads(response.content)["access_token"]

    @property
    def expired(self) -> bool:
        super().expired
        if self._expiration_time is None:  # pragma: no cover
            return False
        return self._expiration_time <= time.time()



import asyncio
import dataclasses
import datetime
import math

from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Collection,
    Iterable,
    Literal,
    Self,
    TypeAlias,
)

from httpx import Response

from aiosalesforce.events import BulkApiBatchConsumptionEvent
from aiosalesforce.utils import json_dumps, json_loads

from ._csv import deserialize_ingest_results, serialize_ingest_data

if TYPE_CHECKING:
    from .client import BulkClientV2

OperationType: TypeAlias = Literal[
    "insert",
    "delete",
    "hardDelete",
    "update",
    "upsert",
]


@dataclasses.dataclass
class JobInfo:
    """Bulk API 2.0 ingest job information."""

    id: str
    operation: str
    object: str
    created_by_id: str
    created_date: datetime.datetime
    system_modstamp: datetime.datetime
    state: Literal[
        "Open",
        "UploadComplete",
        "InProgress",
        "JobComplete",
        "Aborted",
        "Failed",
    ]
    external_id_field_name: str | None
    concurrency_mode: Literal["Parallel"]
    content_type: Literal["CSV"]
    api_version: str
    job_type: Literal["V2Ingest"] | None
    content_url: str
    line_ending: Literal["LF", "CRLF"]
    column_delimiter: Literal[
        "BACKQUOTE",
        "CARET",
        "COMMA",
        "PIPE",
        "SEMICOLON",
        "TAB",
    ]

    @classmethod
    def from_json(cls, data: bytes) -> Self:
        job_info = cls(
            **{
                field.name: (_ := json_loads(data)).get(
                    "".join(
                        [
                            component.capitalize() if i > 0 else component
                            for i, component in enumerate(field.name.split("_"))
                        ]
                    ),
                    None,
                )
                for field in dataclasses.fields(cls)
            }
        )
        for attr in ["created_date", "system_modstamp"]:
            setattr(
                job_info,
                attr,
                datetime.datetime.fromisoformat(getattr(job_info, attr)),
            )
        return job_info


@dataclasses.dataclass
class JobResult:
    """Bulk API 2.0 ingest job result."""

    job_info: JobInfo
    successful_results: list[dict[str, str]]
    failed_results: list[dict[str, str]]
    unprocessed_records: list[dict[str, str]]


class BulkIngestClient:
    """
    Salesforce Bulk API 2.0 ingest client.

    This is a low-level client used to manage ingest jobs.

    Parameters
    ----------
    bulk_client : BulkClientV2
        Bulk API 2.0 client from this client is invoked.

    """

    bulk_client: "BulkClientV2"
    base_url: str
    """Base URL in the format https://[subdomain(s)].my.salesforce.com/services/data/v[version]/jobs/ingest"""

    def __init__(self, bulk_client: "BulkClientV2") -> None:
        self.bulk_client = bulk_client
        self.base_url = "/".join([self.bulk_client.base_url, "ingest"])

    async def create_job(
        self,
        operation: OperationType,
        sobject: str,
        external_id_field: str | None = None,
        assignment_rule_id: str | None = None,
    ) -> JobInfo:
        """
        Create a new ingest job.

        Parameters
        ----------
        operation : {"insert", "delete", "hardDelete", "update", "upsert"}
            Operation to perform.
        sobject : str
            Salesforce object name.
        external_id_field : str | None, optional
            External ID field name, by default None.
            Used for upsert operations.
        assignment_rule_id : str | None, optional
            The ID of an assignment rule to run for a Case or a Lead.
            By default None.

        Returns
        -------
        JobInfo
            _description_
        """
        payload: dict[str, str] = {
            "columnDelimiter": "COMMA",
            "contentType": "CSV",
            "lineEnding": "LF",
            "object": sobject,
            "operation": operation,
        }
        if assignment_rule_id is not None:
            payload["assignmentRuleId"] = assignment_rule_id
        if external_id_field is not None:
            payload["externalIdFieldName"] = external_id_field
        response = await self.bulk_client.salesforce_client.request(
            "POST",
            self.base_url,
            content=json_dumps(payload),
            headers={"Content-Type": "application/json", "Accept": "application/json"},
        )
        return JobInfo.from_json(response.content)

    async def get_job(self, job_id: str) -> JobInfo:
        """
        Get information about ingest job.

        Parameters
        ----------
        job_id : str
            Ingest job ID.

        Returns
        -------
        JobInfo
            Job information.

        """
        response = await self.bulk_client.salesforce_client.request(
            "GET",
            f"{self.base_url}/{job_id}",
            headers={"Accept": "application/json"},
        )
        return JobInfo.from_json(response.content)

    async def list_jobs(
        self,
        is_pk_chunking_enabled: bool | None = None,
    ) -> AsyncIterator[JobInfo]:
        """
        List all ingest jobs.

        Parameters
        ----------
        is_pk_chunking_enabled : bool | None, optional
            Filter by primary key chunking enabled, by default None.

        Yields
        ------
        JobInfo
            Job information.

        """
        params: dict[str, bool] | None = None
        if is_pk_chunking_enabled is not None:
            params = {"isPkChunkingEnabled": is_pk_chunking_enabled}

        next_url: str | None = None
        while True:
            if next_url is None:
                response = await self.bulk_client.salesforce_client.request(
                    "GET",
                    self.base_url,
                    params=params,
                    headers={"Accept": "application/json"},
                )
            else:
                response = await self.bulk_client.salesforce_client.request(
                    "GET",
                    f"{self.bulk_client.salesforce_client.base_url}{next_url}",
                    headers={"Accept": "application/json"},
                )
            response_json: dict = json_loads(response.content)
            for record in response_json["records"]:
                yield JobInfo.from_json(json_dumps(record))
            next_url = response_json.get("nextRecordsUrl", None)
            if next_url is None:
                break

    async def abort_job(self, job_id: str) -> JobInfo:
        """
        Abort ingest job.

        Parameters
        ----------
        job_id : str
            Ingest job ID.

        Returns
        -------
        JobInfo
            Job information.

        """
        response = await self.bulk_client.salesforce_client.request(
            "PATCH",
            f"{self.base_url}/{job_id}",
            content=json_dumps({"state": "Aborted"}),
            headers={"Content-Type": "application/json", "Accept": "application/json"},
        )
        return JobInfo.from_json(response.content)

    async def delete_job(self, job_id: str) -> None:
        """
        Delete ingest job.

        Parameters
        ----------
        job_id : str
            Ingest job ID.

        """
        await self.bulk_client.salesforce_client.request(
            "DELETE",
            f"{self.base_url}/{job_id}",
        )

    async def upload_job_data(
        self,
        job_id: str,
        data: bytes,
    ) -> JobInfo:
        """
        Upload data for an ingest job.

        Job must be in the "Open" state.

        Parameters
        ----------
        job_id : str
            Ingest job ID.
        data : bytes
            CSV data to upload.

        Returns
        -------
        JobInfo
            Job information.

        """
        await self.bulk_client.salesforce_client.request(
            "PUT",
            f"{self.base_url}/{job_id}/batches",
            content=data,
            headers={"Content-Type": "text/csv"},
        )
        response = await self.bulk_client.salesforce_client.request(
            "PATCH",
            f"{self.base_url}/{job_id}",
            content=json_dumps({"state": "UploadComplete"}),
            headers={"Content-Type": "application/json", "Accept": "application/json"},
        )
        await self.bulk_client.salesforce_client.event_bus.publish_event(
            BulkApiBatchConsumptionEvent(
                type="bulk_api_batch_consumption",
                response=response,
                # WARN Bulk API 2.0 does not provide a way to get the number of batches
                #      consumed in a job. Number of batches is estimated based on the
                #      Salesforce docs saying that a separate batch is created for every
                #      10,000 records in data. First row is header and is not counted.
                count=math.ceil((len(data.strip(b"\n").split(b"\n")) - 1) / 10_000),
            )
        )
        return JobInfo.from_json(response.content)

    async def __perform_operation(
        self,
        operation: OperationType,
        sobject: str,
        data: bytes,
        external_id_field: str | None = None,
        assignment_rule_id: str | None = None,
        polling_interval: float = 5.0,
    ) -> JobResult:
        job = await self.create_job(
            operation,
            sobject,
            external_id_field=external_id_field,
            assignment_rule_id=assignment_rule_id,
        )
        job = await self.upload_job_data(job.id, data)
        while job.state.lower().strip(" ") in {"open", "uploadcomplete", "inprogress"}:
            await asyncio.sleep(polling_interval)
            job = await self.get_job(job.id)

        tasks: list[asyncio.Task[Response]] = []
        async with asyncio.TaskGroup() as tg:
            for type_ in [
                "successfulResults",
                "failedResults",
                "unprocessedrecords",
            ]:
                tasks.append(
                    tg.create_task(
                        self.bulk_client.salesforce_client.request(
                            "GET",
                            f"{self.base_url}/{job.id}/{type_}",
                        )
                    )
                )

        return JobResult(
            job_info=job,
            successful_results=deserialize_ingest_results(
                tasks[0].result().content,
            ),
            failed_results=deserialize_ingest_results(
                tasks[1].result().content,
            ),
            unprocessed_records=deserialize_ingest_results(
                tasks[2].result().content,
            ),
        )

    async def perform_operation(
        self,
        operation: OperationType,
        sobject: str,
        data: Iterable[dict[str, Any]],
        fieldnames: Collection[str] | None = None,
        max_size_bytes: int = 100_000_000,
        max_records: int = 150_000_000,
        external_id_field: str | None = None,
        assignment_rule_id: str | None = None,
        polling_interval: float = 5.0,
    ) -> AsyncIterator[JobResult]:
        """
        Perform a bulk ingest operation.

        Parameters
        ----------
        operation : {"insert", "delete", "hardDelete", "update", "upsert"}
            Operation to perform.
        sobject : str
            Salesforce object name.
        data : Iterable[dict[str, Any]]
            Sequence of records to ingest.
        fieldnames : Collection[str], optional
            Field names, determines order of fields in the CSV file.
            By default field names are inferred from the records. This is slow, so
            if you know the field names in advance, it is recommended to provide them.
            If a record is missing a field, it will be written as an empty string.
            If a record has a field not in `fieldnames`, an error will be raised.
        max_size_bytes : int, optional
            Maximum size of each CSV file in bytes.
            The default of 100MB is recommended by Salesforce recommends.
            This accounts for base64 encoding increases in size by up to 50%.
        max_records : int, optional
            Maximum number of records in each CSV file. By default 150,000,000.
            This corresponds to the maximum number of records in a 24-hour period.
        external_id_field : str | None, optional
            External ID field name, by default None.
            Used for upsert operations.
        assignment_rule_id : str | None, optional
            The ID of an assignment rule to run for a Case or a Lead.
            By default None.
        polling_interval : float, optional
            Interval in seconds to poll the job status.
            By default 5.0 seconds.

        Yields
        ------
        JobResult
            Job result containing job information and successful, failed,
            and unprocessed records.

        """
        tasks: list[asyncio.Task[JobResult]] = []
        for csv_payload in serialize_ingest_data(
            data,
            fieldnames=fieldnames,
            max_size_bytes=max_size_bytes,
            max_records=max_records,
        ):
            tasks.append(
                asyncio.create_task(
                    self.__perform_operation(
                        operation,
                        sobject,
                        csv_payload,
                        external_id_field=external_id_field,
                        assignment_rule_id=assignment_rule_id,
                        polling_interval=polling_interval,
                    )
                )
            )
        for future in asyncio.as_completed(tasks):
            yield await future



import asyncio
import random
import time

from typing import TypedDict

from httpx import AsyncClient, Request, Response

from aiosalesforce.events import EventBus, RestApiCallConsumptionEvent, RetryEvent

from .rules import ExceptionRule, ResponseRule


class RetryBase:
    __slots__ = (
        "response_rules",
        "exception_rules",
        "max_retries",
        "timeout",
        "backoff_base",
        "backoff_factor",
        "backoff_max",
        "backoff_jitter",
    )

    response_rules: list[ResponseRule]
    exception_rules: list[ExceptionRule]
    max_retries: int
    timeout: float
    backoff_base: float
    backoff_factor: float
    backoff_max: float
    backoff_jitter: bool

    def __init__(
        self,
        response_rules: list[ResponseRule] | None = None,
        exception_rules: list[ExceptionRule] | None = None,
        max_retries: int = 3,
        timeout: float = 60,
        backoff_base: float = 0.5,
        backoff_factor: float = 2,
        backoff_max: float = 10,
        backoff_jitter: bool = True,
    ) -> None:
        self.response_rules = response_rules or []
        self.exception_rules = exception_rules or []
        self.max_retries = max_retries
        self.timeout = timeout
        self.backoff_base = backoff_base
        self.backoff_factor = backoff_factor
        self.backoff_max = backoff_max
        self.backoff_jitter = backoff_jitter


class RetryCounts(TypedDict):
    total: int
    exception: dict[ExceptionRule, int]
    response: dict[ResponseRule, int]


class RetryContext(RetryBase):
    """Context for handling retries for a single request."""

    __slots__ = ("start", "retry_count")

    start: float
    retry_count: RetryCounts

    def __init__(
        self,
        response_rules: list[ResponseRule],
        exception_rules: list[ExceptionRule],
        max_retries: int,
        timeout: float,
        backoff_base: float,
        backoff_factor: float,
        backoff_max: float,
        backoff_jitter: bool,
    ) -> None:
        super().__init__(
            response_rules=response_rules,
            exception_rules=exception_rules,
            max_retries=max_retries,
            timeout=timeout,
            backoff_base=backoff_base,
            backoff_factor=backoff_factor,
            backoff_max=backoff_max,
            backoff_jitter=backoff_jitter,
        )

        self.start = time.time()
        self.retry_count = {
            "total": 0,
            "response": {rule: 0 for rule in self.response_rules},
            "exception": {rule: 0 for rule in self.exception_rules},
        }

    async def send_request_with_retries(
        self,
        httpx_client: AsyncClient,
        event_bus: EventBus,
        semaphore: asyncio.Semaphore,
        request: Request,
    ) -> Response:
        """
        Send a request and retry it if necessary in accordance with the retry policy.

        Does not guarantee that the returned response is OK (status code < 300),
        only that the request was retried according to the policy.

        Emits the following events:\n
        * RetryEvent if the request is retried
        * RestApiCallConsumptionEvent for each request
            that did not raise an exception

        Parameters
        ----------
        httpx_client : AsyncClient
            HTTP client to send the request.
        event_bus : EventBus
            Event bus to publish events.
        semaphore : asyncio.Semaphore
            Semaphore to limit the number of simultaneous requests.
        request : Request
            Request to send.

        Returns
        -------
        Response
            Response from the request.
            Not guaranteed to be OK (status code < 300).

        """
        while True:
            try:
                async with semaphore:
                    response = await httpx_client.send(request)
            except Exception as exc:
                if await self.should_retry(exc):
                    await asyncio.gather(
                        self.sleep(),
                        event_bus.publish_event(
                            RetryEvent(
                                type="retry",
                                attempt=self.retry_count["total"],
                                request=request,
                                exception=exc,
                            )
                        ),
                    )
                    continue
                raise
            await event_bus.publish_event(
                RestApiCallConsumptionEvent(
                    type="rest_api_call_consumption",
                    response=response,
                    count=1,
                )
            )
            if not response.is_success and await self.should_retry(response):
                await asyncio.gather(
                    self.sleep(),
                    event_bus.publish_event(
                        RetryEvent(
                            type="retry",
                            attempt=self.retry_count["total"],
                            request=request,
                            response=response,
                        )
                    ),
                )
                continue
            return response

    async def should_retry(self, value: Response | Exception) -> bool:
        """
        Determine if the request should be retried.

        If the request should be retried, the total retry count and the retry count
        for the rule responsible for the retry are incremented.

        Parameters
        ----------
        value : Response | Exception
            Response or Exception from the request.

        Returns
        -------
        bool
            True if the request should be retried, False otherwise.

        """
        if (
            self.retry_count["total"] >= self.max_retries
            or time.time() - self.start > self.timeout
        ):
            return False
        condition: bool = False
        match value:
            case Response():
                condition = await self.__evaluate_response_rules(value)
            case Exception():
                condition = await self.__evaluate_exception_rules(value)
            case _:  # pragma: no cover
                raise TypeError("Value must be a Response or an Exception")
        if condition:
            self.retry_count["total"] += 1
        return condition

    async def __evaluate_response_rules(self, response: Response) -> bool:
        for rule in self.response_rules:
            if self.retry_count["response"][
                rule
            ] < rule.max_retries and await rule.should_retry(response):
                self.retry_count["response"][rule] += 1
                return True
        return False

    async def __evaluate_exception_rules(self, exception: Exception) -> bool:
        for rule in self.exception_rules:
            if self.retry_count["exception"][rule] >= rule.max_retries:
                return False
            if await rule.should_retry(exception):
                self.retry_count["exception"][rule] += 1
                return True
        return False

    async def sleep(self) -> None:
        """Sleep between retries based on the backoff policy."""
        # (total - 1) because this is called after incrementing the total count
        sleep_time = min(
            self.backoff_base
            * (self.backoff_factor ** (self.retry_count["total"] - 1)),
            self.backoff_max,
        )
        if self.backoff_jitter:
            sleep_time = random.uniform(0, sleep_time)  # noqa: S311
        await asyncio.sleep(sleep_time)


class RetryPolicy(RetryBase):
    """
    Policy for retrying requests.

    Parameters
    ----------
    response_rules : list[ResponseRule], optional
        Rules for retrying requests based on their responses.
    exception_rules : list[ExceptionRule], optional
        Rules for retrying requests after an exception.
    max_retries : int, optional
        Maximum total number of retries. By default 3.
    timeout : float, optional
        Maximum time to retry. By default 60 seconds.
        Timeout is best effort and may exceed the specified time
        by up to `backoff_max` seconds.
    backoff_base : float, optional
        Base time to sleep between retries. By default 0.5 seconds.
    backoff_factor : float, optional
        Factor to increase sleep time between retries. By default 2.
    backoff_max : float, optional
        Maximum time to sleep between retries. By default 10 seconds.
    backoff_jitter : bool, optional
        If True, adds jitter to sleep time. By default True.

    """

    def create_context(self) -> RetryContext:
        """
        Create a new retry context.

        Retry context is used to handle retries for a single request.

        """
        return RetryContext(
            response_rules=self.response_rules,
            exception_rules=self.exception_rules,
            max_retries=self.max_retries,
            timeout=self.timeout,
            backoff_base=self.backoff_base,
            backoff_factor=self.backoff_factor,
            backoff_max=self.backoff_max,
            backoff_jitter=self.backoff_jitter,
        )



import dataclasses
import logging

from functools import cached_property
from typing import TYPE_CHECKING, Any, Iterable

if TYPE_CHECKING:
    from aiosalesforce.client import Salesforce

from .ingest import BulkIngestClient, JobInfo, OperationType

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class IngestResult:
    """Bulk API 2.0 ingest operation result."""

    jobs: list[JobInfo]
    successful_results: list[dict[str, str]]
    failed_results: list[dict[str, str]]
    unprocessed_records: list[dict[str, str]]


class BulkClientV2:
    """
    Salesforce Bulk API 2.0 client.

    Use this client to execute bulk ingest and query operations.

    Parameters
    ----------
    salesforce_client : Salesforce
        Salesforce client.

    """

    salesforce_client: "Salesforce"
    base_url: str
    """Base URL in the format https://[subdomain(s)].my.salesforce.com/services/data/v[version]/jobs"""

    def __init__(self, salesforce_client: "Salesforce") -> None:
        self.salesforce_client = salesforce_client
        self.base_url = "/".join(
            [
                f"{self.salesforce_client.base_url}",
                "services",
                "data",
                f"v{self.salesforce_client.version}",
                "jobs",
            ]
        )

    @cached_property
    def ingest(self) -> BulkIngestClient:
        """Manage ingest jobs at a low level."""
        return BulkIngestClient(self)

    async def __perform_operation(
        self,
        operation: OperationType,
        sobject: str,
        data: Iterable[dict[str, Any]],
        external_id_field: str | None = None,
        assignment_rule_id: str | None = None,
    ) -> IngestResult:
        result = IngestResult([], [], [], [])
        async for job_result in self.ingest.perform_operation(
            operation=operation,
            sobject=sobject,
            data=data,
            external_id_field=external_id_field,
            assignment_rule_id=assignment_rule_id,
        ):
            result.jobs.append(job_result.job_info)
            result.successful_results.extend(job_result.successful_results)
            result.failed_results.extend(job_result.failed_results)
            result.unprocessed_records.extend(job_result.unprocessed_records)
        return result

    async def insert(
        self,
        sobject: str,
        data: Iterable[dict[str, Any]],
        assignment_rule_id: str | None = None,
    ) -> IngestResult:
        """
        Create new records in Salesforce.

        Parameters
        ----------
        sobject : str
            Salesforce object name.
        data : Iterable[dict[str, Any]]
            Records to create.
        assignment_rule_id : str | None, default None
            The ID of an assignment rule to run for a Case or a Lead.

        Returns
        -------
        IngestResult
            Bulk API 2.0 ingest job results.

        """
        return await self.__perform_operation(
            "insert",
            sobject=sobject,
            data=data,
            assignment_rule_id=assignment_rule_id,
        )

    async def update(
        self,
        sobject: str,
        data: Iterable[dict[str, Any]],
        assignment_rule_id: str | None = None,
    ) -> IngestResult:
        """
        Update existing records in Salesforce.

        Parameters
        ----------
        sobject : str
            Salesforce object name.
        data : Iterable[dict[str, Any]]
            Records to update.
        assignment_rule_id : str | None, default None
            The ID of an assignment rule to run for a Case or a Lead.

        Returns
        -------
        IngestResult
            Bulk API 2.0 ingest job results.

        """
        return await self.__perform_operation(
            "update",
            sobject=sobject,
            data=data,
            assignment_rule_id=assignment_rule_id,
        )

    async def upsert(
        self,
        sobject: str,
        data: Iterable[dict[str, Any]],
        external_id_field: str,
        assignment_rule_id: str | None = None,
    ) -> IngestResult:
        """
        Create or update records in Salesforce.

        Parameters
        ----------
        sobject : str
            Salesforce object name.
        data : Iterable[dict[str, Any]]
            Records to create or update.
        external_id_field : str
            External ID field name.
        assignment_rule_id : str | None, default None
            The ID of an assignment rule to run for a Case or a Lead.

        Returns
        -------
        IngestResult
            Bulk API 2.0 ingest job results.

        """
        return await self.__perform_operation(
            "upsert",
            sobject=sobject,
            data=data,
            external_id_field=external_id_field,
            assignment_rule_id=assignment_rule_id,
        )

    async def delete(
        self,
        sobject: str,
        data: Iterable[dict[str, Any]],
        hard: bool = False,
    ) -> IngestResult:
        """
        Delete records from Salesforce.

        Parameters
        ----------
        sobject : str
            Salesforce object name.
        data : Iterable[dict[str, Any]]
            Records to delete.
        hard : bool, default False
            Whether to hard delete records.

        Returns
        -------
        IngestResult
            Bulk API 2.0 ingest job results.

        """
        return await self.__perform_operation(
            "hardDelete" if hard else "delete",
            sobject=sobject,
            data=data,
            assignment_rule_id=None,
        )



import time

from typing import TYPE_CHECKING

from aiosalesforce.auth.base import Auth
from aiosalesforce.events import (
    RequestEvent,
    ResponseEvent,
)
from aiosalesforce.exceptions import AuthenticationError
from aiosalesforce.utils import json_loads

if TYPE_CHECKING:
    from aiosalesforce.client import Salesforce


class ClientCredentialsFlow(Auth):
    """
    Authenticate using the OAuth 2.0 Client Credentials Flow.

    https://help.salesforce.com/s/articleView?id=sf.remoteaccess_oauth_client_credentials_flow.htm&type=5

    Parameters
    ----------
    client_id : str
        Client ID.
    client_secret : str
        Client secret.
    timeout : float, optional
        Timeout for the access token in seconds.
        By default assumed to never expire.

    """

    def __init__(
        self,
        client_id: str,
        client_secret: str,
        timeout: float | None = None,
    ) -> None:
        super().__init__()
        self.client_id = client_id
        self.client_secret = client_secret
        self.timeout = timeout

        self._expiration_time: float | None = None

    async def _acquire_new_access_token(self, client: "Salesforce") -> str:
        request = client.httpx_client.build_request(
            "POST",
            f"{client.base_url}/services/oauth2/token",
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "Accept": "application/json",
            },
            data={
                "grant_type": "client_credentials",
                "client_id": self.client_id,
                "client_secret": self.client_secret,
            },
        )
        await client.event_bus.publish_event(
            RequestEvent(
                type="request",
                request=request,
            )
        )
        retry_context = client.retry_policy.create_context()
        response = await retry_context.send_request_with_retries(
            httpx_client=client.httpx_client,
            event_bus=client.event_bus,
            semaphore=client._semaphore,
            request=request,
        )
        if not response.is_success:
            try:
                response_json = json_loads(response.content)
                error_code = response_json["error"]
                error_message = response_json["error_description"]
            except Exception:  # pragma: no cover
                error_code = None
                error_message = response.text
            raise AuthenticationError(
                f"[{error_code}] {error_message}" if error_code else error_message,
                response=response,
                error_code=error_code,
                error_message=error_message,
            )
        await client.event_bus.publish_event(
            ResponseEvent(
                type="response",
                response=response,
            )
        )
        if self.timeout is not None:
            self._expiration_time = time.time() + self.timeout
        return json_loads(response.content)["access_token"]

    @property
    def expired(self) -> bool:
        super().expired
        if self._expiration_time is None:  # pragma: no cover
            return False
        return self._expiration_time <= time.time()



import datetime
import string

from typing import Any, LiteralString, TypeAlias

import orjson


def json_dumps(data: str | bytes | bytearray | Any) -> bytes:
    """
    Serialize data to a JSON formatted bytes object.

    Utility function used to allow users to pass an already serialized JSON.

    Parameters
    ----------
    data : str | bytes | bytearray | Any
        Data to be serialized.

    Returns
    -------
    bytes
        JSON formatted bytes object.

    """
    if isinstance(data, str):
        return data.encode("utf-8")
    if isinstance(data, bytearray):
        return bytes(data)
    if isinstance(data, bytes):
        return data
    return orjson.dumps(data, option=orjson.OPT_OMIT_MICROSECONDS | orjson.OPT_UTC_Z)


def json_loads(data: str | bytes | bytearray) -> Any:
    """
    Deserialize JSON formatted data.

    Parameters
    ----------
    data : str | bytes | bytearray
        JSON formatted data.

    Returns
    -------
    Any
        Deserialized data.

    """
    return orjson.loads(data)


# https://developer.salesforce.com/docs/atlas.en-us.soql_sosl.meta/soql_sosl/sforce_api_calls_soql_select_quotedstringescapes.htm
SOQL_RESERVED_CHARACTERS = {
    "\\": "\\\\",  # Backslash
    '"': '\\"',  # Double quote
    "'": "\\'",  # Single quote
}
SOQL_ESCAPE_SEQUENCES = {
    "\n": "\\n",  # Newline
    "\r": "\\r",  # Carriage return
    "\t": "\\t",  # Tab
    "\b": "\\b",  # Backspace (called 'Bell' in SOQL documentation)
    "\f": "\\f",  # Form feed
}
SOQL_ESCAPE_SEQUENCES_FOR_LIKE = {
    "_": "\\_",  # Underscore
    "%": "\\%",  # Percent
}

_ScalarSoqlParameter: TypeAlias = (
    bool | None | int | float | datetime.datetime | datetime.date | str
)
_SoqlParameter: TypeAlias = (
    _ScalarSoqlParameter
    | list[_ScalarSoqlParameter]
    | tuple[_ScalarSoqlParameter]
    | set[_ScalarSoqlParameter]
)


def _sanitize_soql_query_parameter(  # noqa: PLR0911
    value: _SoqlParameter,
    like: bool = False,
) -> str:
    """
    Sanitize SOQL query parameter.

    Unless 'like' is True, the returned string is enclosed in single quotes.

    Parameters
    ----------
    value : str
        Value to be sanitized.
    like : bool, default True
        Whether the string is to be used in a LIKE clause, by default False.
        When True, % and _ characters are escaped and
        the returned string is not enclosed in single quotes.

    Returns
    -------
    str
        Sanitized string.

    """
    match value:
        case True:
            return "true"
        case False:
            return "false"
        case None:
            return "null"
        case int() | float():
            return str(value)
        case datetime.datetime():
            # yyyy-MM-ddTHH:mm:ss.SSS+/-HH:mm
            return value.isoformat(timespec="milliseconds")
        case datetime.date():
            return value.strftime(r"%Y-%m-%d")
        case str():
            pass
        case list() | tuple() | set():
            sanitized_values = []
            for item in value:
                if isinstance(item, (list, tuple, set)):
                    raise TypeError("Nested lists, tuples, and sets are not allowed")
                sanitized_values.append(_sanitize_soql_query_parameter(item, like=like))
            return f"({','.join(sanitized_values)})"
        case _:
            raise TypeError(f"Invalid SOQL parameter type: {type(value).__name__}")

    sanitized_value = value
    for char, sanitized_char in SOQL_RESERVED_CHARACTERS.items():
        sanitized_value = sanitized_value.replace(char, sanitized_char)
    for char, sanitized_char in SOQL_ESCAPE_SEQUENCES.items():
        sanitized_value = sanitized_value.replace(char, sanitized_char)
    if like:
        for char, sanitized_char in SOQL_ESCAPE_SEQUENCES_FOR_LIKE.items():
            sanitized_value = sanitized_value.replace(char, sanitized_char)
    return sanitized_value if like else f"'{sanitized_value}'"


class SoqlFormatter(string.Formatter):
    def format_field(self, value: Any, format_spec: str) -> Any:
        match format_spec:
            case "like":
                return super().format_field(
                    _sanitize_soql_query_parameter(value, like=True),
                    "",
                )
            case _:
                return super().format_field(
                    _sanitize_soql_query_parameter(value, like=False),
                    format_spec,
                )


def format_soql(query: LiteralString, *args, **kwargs) -> str:
    """
    Format SOQL query template with dynamic parameters.

    While SOQL queries are safe by design (cannot create/update/delete records),
    it is still possible to expose sensitive information via SOQL injection.
    It is always recommended to use this function instead of string formatting.

    You should never surround your parameters with single quotes in the query
    template - this is done automatically by this function when necessary.
    E.g., "SELECT Id FROM Object WHERE Value = {value}"

    The only exception to this rule is when you use the 'like' format spec.
    The 'like' format spec is used to escape special characters in a LIKE pattern
    when a portion of it is dynamic.
    E.g., "SELECT Id FROM Object WHERE Value LIKE '%{value:like}'"

    If you don't use the 'like' format spec when formatting LIKE statements,
    you will get unnecessary quotes and special query characters
    (% and _) will not be escaped. This would make you vulnerable to SOQL injection.

    Examples
    --------
    >>> format_soql("SELECT Id FROM Account WHERE Name = {}", "John Doe")
    "SELECT Id FROM Account WHERE Name = 'John Doe'"
    >>> format_soql("SELECT Id FROM Account WHERE Name = {name}", name="John Doe")
    "SELECT Id FROM Account WHERE Name = 'John Doe'"
    >>> format_soql("SELECT Id FROM Account WHERE Name LIKE {value}", value="John%")
    "SELECT Id FROM Account WHERE Name LIKE 'John%'"
    >>> format_soql(
    >>>     "SELECT Id FROM Record WHERE Description LIKE '% fails {pattern:like}'",
    >>>     pattern="50% of the time",
    >>> )
    "SELECT Id FROM Record WHERE Description LIKE '% fails 50\\% of the time'"

    Parameters
    ----------
    query : LiteralString
        SOQL query template.

    Returns
    -------
    str
        Formatted SOQL query.

    """
    # Use vformat to avoid unnecessary args/kwargs unpacking
    return SoqlFormatter().vformat(query, args, kwargs)
